# -*- coding: utf-8 -*-
"""
Created on Thu Jan  2 10:41:18 2020

@author: icruicks
"""

import numpy as np
from sknetwork.clustering import Louvain, BiLouvain
from sklearn.cluster import KMeans
from sklearn.preprocessing import normalize
from scipy.sparse import csr_matrix, lil_matrix
from scipy.spatial.distance import pdist, squareform
from Leiden_Clustering import Leiden_Multiplex


class Cross_Learning:
    
    def __init__(self, alpha=0.5, model='additive', cluster_usage='alternate',
                 cluster_association ='similarity', graph_weighting ='straight', 
                 iterations = 5):
        
        self.alpha = alpha
        self.model= model
        self.cluster_usage = cluster_usage
        self.cluster_association = cluster_association
        self.graph_weighting = graph_weighting
        self.iterations = iterations
    
    def fit_predict(self, base_clusters, modal_graphs):
        ba_matrices =[]
        
        for base_cluster in base_clusters:
            ba_matrix = self._produce_ba_matrix(base_cluster)
            ba_matrices.append(ba_matrix)
        
        C_t = ba_matrices.copy()
        graph_cluster = Louvain()
        converged = False
        self.iterate = 0
        if self.graph_weighting =='straight':
            self.graphs = [graph.copy() for graph in modal_graphs]
        elif self.graph_weighting =='neighbors':
            neighbors_weighted = [np.nan_to_num(squareform(1-pdist(graph.todense(), metric='cosine'))) for graph in modal_graphs]
            self.graphs = [csr_matrix(graph) for graph in neighbors_weighted]
        
        while not converged:
            C_new = []
            if self.cluster_usage =="all":
                C_prime = np.concatenate(C_t, axis=1)
                # self.tau = 1-np.mean([len(G.data)/(G.shape[0]*G.shap1[1]) for G in self.graphs])
                if self.cluster_association =='similarity':
                    U_prime = C_prime @ C_prime.T
                    U_prime[U_prime < len(self.graphs)] =0
                    U_prime[U_prime > 0] =1
                elif self.cluster_association =='dissimilarity':
                    CD = squareform(pdist(C_prime, metric='jaccard'))
                    U_prime = squareform(1-pdist(CD, metric='cosine'))
                    # U_prime = self._prune(U_prime)
            
            
            for m in range(len(modal_graphs)):
                if self.cluster_usage =="alternate":
                    C_prime = np.concatenate([C_t[mode] for mode in range(len(self.graphs)) if mode !=m], axis=1)
                    if self.cluster_association =='similarity':
                        U_prime = 1/(len(self.graphs)-1) * (C_prime @ C_prime.T)
                    elif self.cluster_association =='dissimilarity':
                        CD = squareform(pdist(C_prime, metric='jaccard'))
                        U_prime = squareform(1-pdist(CD, metric='cosine'))
                    # self.tau = 1-len(self.graphs[m].data)/(self.graphs[m].shape[0] * self.graphs[m].shape[1])
                    # U_prime = self._prune(U_prime)
                
                if self.model == 'additive':
                    A_t = csr_matrix(self.alpha * self.graphs[m] + (1-self.alpha) * U_prime)
                elif self.model == 'multiplicative':
                    A_t = csr_matrix(self.graphs[m].multiply(U_prime))
                elif self.model == 'augmented':
                    aux = np.concatenate([C_prime, np.zeros([C_prime.shape[1],C_prime.shape[1]])], axis=0)
                    A_t = np.concatenate([self.graphs[m].todense(), C_prime.T], axis=0)
                    A_t = csr_matrix(np.concatenate([A_t, aux], axis=1))
                elif self.model =='diffusion':
                    kernel = normalize(self.graphs[m], norm='l1', axis=1, copy=False)
                    full = normalize(U_prime)
                    for _ in range(3):
                        A_t = csr_matrix(self.alpha * kernel @ full @ kernel.T + (1-self.alpha)*np.eye(U_prime.shape[0]))
                        full = A_t.copy()
                else:
                    raise ValueError(self.model+" not a valid model for combining graph and side information")
                
                if self.model == 'augmented':
                    all_clusters = graph_cluster.fit(A_t).labels_
                    modal_clusters = all_clusters[:C_prime.shape[0]]
                else:
                    modal_clusters = graph_cluster.fit(A_t).labels_
                BA_m_t = self._produce_ba_matrix(modal_clusters)
                # CA_m_t = BA_m_t @ BA_m_t.T
                # new_graph = lil_matrix(self.graphs[m])
                # new_graph[(1-CA_m_t).nonzero()] = new_graph[(1-CA_m_t).nonzero()] * 0.8
                # self.graphs[m] = csr_matrix(new_graph)
                C_new.append(BA_m_t)
                
            if self.iterate >= self.iterations:
                converged = True
            else:
                self.iterate +=1
                C_t = C_new.copy()
            
            # CA_t = np.concatenate(C_t, axis=1)
            # CA_new = np.concatenate(C_new, axis=1)
            # U_t = CA_t @ CA_t.T
            # U_new = CA_new @ CA_new.T
            # obj_val = np.sum(np.abs(U_t-U_new))
            # if np.abs(obj_val-obj_val_prev) >0:
            #     self.iterations +=1
            #     obj_val_prev = obj_val
            #     C_t = C_new.copy()
            # else:
            #     if self.iterations >1:
            #         converged = True
            #         self.final_obj_val = obj_val_prev
                    
            #     self.iterations +=1
                      
        CA_final = np.concatenate(C_t, axis=1)
        return BiLouvain().fit(CA_final).row_labels_
    
    def _produce_ba_matrix(self, cluster):
        ba_matrix = np.zeros((cluster.size, cluster.max()+1))
        ba_matrix[np.arange(cluster.size), cluster] = 1
        return ba_matrix
    
    def _prune(self, adjacency):
        thresh = np.apply_along_axis(lambda row: np.quantile(row[np.nonzero(row)], self.tau), 1, adjacency)
        mins = np.apply_along_axis(lambda row: np.min(row[np.nonzero(row)]), 1, adjacency)
        global_thresh = np.quantile(adjacency[np.nonzero(adjacency)], self.tau)
        thresh[mins > global_thresh] = 0
        adjacency[adjacency < thresh] = 0
        # adjacency[adjacency < global_thresh] = 0
        return adjacency
        
    
    
class Spectral:
    
    def __init__(self, k=10, alpha=0.5, model='additive', cluster_usage='alternate',
                 cluster_association ='similarity', graph_weighting ='straight'):
        
        self.k = k
        self.alpha = alpha
        self.model = model
        self.cluster_usage = cluster_usage
        self.cluster_association = cluster_association
        self.graph_weighting = graph_weighting
    
    def fit_predict(self, base_clusters, modal_graphs):
        
        if self.graph_weighting =='straight':
            self.graphs = [graph.copy() for graph in modal_graphs]
        elif self.graph_weighting =='neighbors':
            neighbors_weighted = [squareform(1-pdist(graph.todense(), metric='cosine')) for graph in modal_graphs]
            self.graphs = [csr_matrix(graph) for graph in neighbors_weighted]
        
        ba_matrices =[]
        idxs = []
        idxs_iter = 0
        
        for base_cluster in base_clusters:
            ba_matrix = np.zeros((base_cluster.size, base_cluster.max()+1))
            ba_matrix[np.arange(base_cluster.size), base_cluster] = 1
            ba_matrices.append(ba_matrix)
            idxs.append([idxs_iter, idxs_iter+ba_matrix.shape[1]])
            idxs_iter += ba_matrix.shape[1]

        eigenvectors_prime =[]
        if self.cluster_usage =="all":
            B = np.concatenate(ba_matrices, axis=1)
            self.tau = 1-np.mean([len(G.data)/(G.shape[0]*G.shap1[1]) for G in self.graphs])
            if self.cluster_association =='similarity':
                U_prime = B @ B.T
                U_prime[U_prime < len(modal_graphs)] =0
                U_prime[U_prime > 0] =1
            elif self.cluster_association =='dissimilarity':
                CD = squareform(pdist(B, metric='jaccard'))
                U_prime = squareform(1-pdist(CD, metric='cosine'))
                if self.model == 'additive':
                    U_prime = self._prune(U_prime)
            
        for i in range(len(modal_graphs)):
            G = self.graphs[i]
            
            if self.cluster_usage =="alternate":
                B = np.concatenate([ba_matrices[j] for j in range(len(modal_graphs)) if j!=i], axis=1)
                if self.cluster_association =='similarity':
                    U_prime = 1/(len(modal_graphs)-1)* B @ B.T
                elif self.cluster_association =='dissimilarity':
                    CD = squareform(pdist(B, metric='jaccard'))
                    U_prime = squareform(1-pdist(CD, metric='cosine'))
                if self.model == 'additive':
                    self.tau = 1-len(self.graphs[i].data)/(self.graphs[i].shape[0] * self.graphs[i].shape[1])
                    U_prime = self._prune(U_prime)
            
            if self.model == 'additive':
                G_prime = csr_matrix(self.alpha * G + (1-self.alpha) * U_prime)
            elif self.model == 'multiplicative':
                G_prime = csr_matrix(G.multiply(U_prime))
            elif self.model =='regularized':
                G_prime =csr_matrix(G)
            elif self.model == 'augmented':
                aux = np.concatenate([B, np.zeros([B.shape[1], B.shape[1]])], axis=0)
                G_prime = np.concatenate([G.todense(), B.T], axis=0)
                G_prime = csr_matrix(np.concatenate([G_prime, aux], axis=1))
            else:
                raise ValueError(self.model+" not a valid model for combining graph and side information")
            
            G_prime.setdiag(0)
            # D= np.diag(np.ravel(np.sum(G_prime, axis=1)))
            # D_inv = np.nan_to_num(np.diag(np.ravel(np.sum(G_prime, axis=1))**-0.5))
            # L = np.nan_to_num(D_inv @ (D-G_prime) @ D_inv)
            D_inv = np.nan_to_num(np.diag(np.ravel(1/np.sum(G_prime, axis=1))))
            L = np.nan_to_num(np.eye(D_inv.shape[0])-D_inv @ G_prime)
            if self.model =='regularized':
                _, v =np.linalg.eig(L + self.alpha*U_prime)
            else:
                _, v =np.linalg.eigh(L)
            if self.model == 'augmented':
                eigenvectors_prime.append(np.real(v)[:B.shape[0],:self.k])
            else:
                eigenvectors_prime.append(np.real(v)[:,:self.k])
            
            
        U = np.concatenate(eigenvectors_prime, axis=1)
        c, s, vh = np.linalg.svd(U)
        c = c[:,:self.k]
        clstr = KMeans(n_clusters = self.k)
        
        return clstr.fit_predict(c)
    
    def _prune(self, adjacency):
        thresh = np.apply_along_axis(lambda row: np.quantile(row[np.nonzero(row)], self.tau), 1, adjacency)
        mins = np.apply_along_axis(lambda row: np.min(row[np.nonzero(row)]), 1, adjacency)
        global_thresh = np.quantile(adjacency[np.nonzero(adjacency)], self.tau)
        thresh[mins > global_thresh] = 0
        adjacency[adjacency < thresh] = 0
        # adjacency[adjacency < global_thresh] = 0
        return adjacency
    
    
class Multi_Layer:
    
    def __init__(self, alpha=0.5, model='additive', cluster_usage='alternate',
                 cluster_association ='similarity', graph_weighting ='straight',
                 n_iterations = -1):
        
        self.alpha = alpha
        self.model = model
        self.cluster_usage = cluster_usage
        self.cluster_association = cluster_association
        self.graph_weighting = graph_weighting
        self.n_iterations = n_iterations
        
    def fit_predict(self, base_clusters, modal_graphs):
    
        if self.graph_weighting =='straight':
            self.graphs = [graph.copy() for graph in modal_graphs]
        elif self.graph_weighting =='neighbors':
            neighbors_weighted = [squareform(1-pdist(graph.todense(), metric='cosine')) for graph in modal_graphs]
            self.graphs = [csr_matrix(graph) for graph in neighbors_weighted]
        
        ba_matrices =[]
        for base_cluster in base_clusters:
            ba_matrix = self._produce_ba_matrix(base_cluster)
            ba_matrices.append(ba_matrix)
            
        G_prime = []
        if self.cluster_usage =="all":
            B = np.concatenate(ba_matrices, axis=1)
            self.tau = 1-np.mean([len(G.data)/(G.shape[0]*G.shap1[1]) for G in self.graphs])
            if self.cluster_association =='similarity':
                U_prime = B @ B.T
                U_prime[U_prime < len(modal_graphs)] =0
                U_prime[U_prime > 0] =1
            elif self.cluster_association =='dissimilarity':
                CD = squareform(pdist(B, metric='jaccard'))
                U_prime = squareform(1-pdist(CD, metric='cosine'))
                if self.model == 'additive':
                    U_prime = self._prune(U_prime)
            
        for i in range(len(modal_graphs)):
            G = self.graphs[i]
            
            if self.cluster_usage =="alternate":
                B = np.concatenate([ba_matrices[j] for j in range(len(modal_graphs)) if j!=i], axis=1)
                if self.cluster_association =='similarity':
                    U_prime = 1/(len(modal_graphs)-1)* B @ B.T
                elif self.cluster_association =='dissimilarity':
                    CD = squareform(pdist(B, metric='jaccard'))
                    U_prime = squareform(1-pdist(CD, metric='cosine'))
                if self.model == 'additive':
                    self.tau = 1-len(self.graphs[i].data)/(self.graphs[i].shape[0] * self.graphs[i].shape[1])
                    U_prime = self._prune(U_prime)
            
            if self.model == 'additive':
                G_prime.append(csr_matrix(self.alpha * G + (1-self.alpha) * U_prime))
            elif self.model == 'multiplicative':
                G_prime.append(csr_matrix(G.multiply(U_prime)))
            elif self.model == 'augmented':
                aux = np.concatenate([B, np.zeros([B.shape[1], B.shape[1]])], axis=0)
                G_temp = np.concatenate([G.todense(), B.T], axis=0)
                G_prime.append(csr_matrix(np.concatenate([G_temp, aux], axis=1)))
            else:
                raise ValueError(self.model+" not a valid model for combining graph and side information")
            
        self.transformed_graphs = G_prime.copy()
        lc_clstr = Leiden_Multiplex(n_iterations = self.n_iterations)
        if self.model == 'augmented':
            all_clusters = lc_clstr.fit_transform(G_prime)
            modal_clusters = all_clusters[:B.shape[0]]
        else:
            modal_clusters = lc_clstr.fit_transform(G_prime)
        self.modularities = lc_clstr.modularities
        return modal_clusters
    
    def _produce_ba_matrix(self, cluster):
        ba_matrix = np.zeros((cluster.size, cluster.max()+1))
        ba_matrix[np.arange(cluster.size), cluster] = 1
        return ba_matrix
            
    
    def _prune(self, adjacency):
        thresh = np.apply_along_axis(lambda row: np.quantile(row[np.nonzero(row)], self.tau), 1, adjacency)
        mins = np.apply_along_axis(lambda row: np.min(row[np.nonzero(row)]), 1, adjacency)
        global_thresh = np.quantile(adjacency[np.nonzero(adjacency)], self.tau)
        thresh[mins > global_thresh] = 0
        adjacency[adjacency < thresh] = 0
        # adjacency[adjacency < global_thresh] = 0
        return adjacency
    
    
    
    
    
    
    
    
    
    
    